/* Copyright (c) 2000-2008 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

if (typeof(ko) == 'undefined') {
    ko = {};
}
if (typeof(ko.formatters) == 'undefined') {
    ko.formatters = {};
}

/* Create the ko.formatters scope. */
(function() {

    /************************************************************
     *        Globals                                           *
     ************************************************************/

    var log = ko.logging.getLogger('ko.formatters');
    //log.setLevel(ko.logging.LOG_DEBUG);
    const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";


    /************************************************************
     *         Controller for the formatter commands            *
     ************************************************************/

    // Hide this controller from our global namespace, we simply dont need to
    // touch anything in it from the outside world.
    function FormattersController() {
    }

    // The following two lines ensure proper inheritance (see Flanagan, p. 144).
    FormattersController.prototype = new Controller();
    FormattersController.prototype.constructor = FormattersController;

    FormattersController.prototype.destructor = function() {
    }

    FormattersController.prototype.is_cmd_format_enabled = function () {
        var view = ko.views.manager.currentView;
        return view && view.getAttribute('type') == 'editor';
    };

    FormattersController.prototype.do_cmd_format = function () {
        ko.formatters.formatView(ko.views.manager.currentView);
    };

    //FormattersController.prototype.is_cmd_formatSelection_enabled = function () {
    //    var view = ko.views.manager.currentView;
    //    if (view && view.getAttribute('type') == 'editor') {
    //        var scimoz = view.scintilla.scimoz;
    //        if (scimoz && scimoz.selText) {
    //            document.getElementById("editor_contextmenu_formatters_menu").removeAttribute("collapsed");
    //            return true;
    //        }
    //    }
    //    document.getElementById("editor_contextmenu_formatters_menu").setAttribute("collapsed", "true");
    //    return false;
    //};
    //FormattersController.prototype.do_cmd_formatSelection = function (event) {
    //    ko.formatters.formatView(ko.views.manager.currentView,
    //                             true /* formatSelection */);
    //};


    /************************************************************
     *        Internal private functions                        *
     ************************************************************/

    /**
     * Format the supplied view with the provided formatter and context.
     
     * @param view {Components.interfaces.koIScintillaView}  The view to format.
     * @param formatter {Components.interfaces.koIFormatter}  Formatter to use.
     * @param context {Components.interfaces.koIFormatterContext}  Context to use.
     */
    function _formatViewWithFormatterAndContext(view, formatter, context) {
        try {
            // perLineWhitespace - how much whitespace needs to be added to
            //                     each formatted line.
            var perLineWhitespace = "";
            // firstLineWhitespace - how much whitespace was already in the
            //                          first line.
            var firstLineWhitespace = "";
            var preceedingSelectionWhitespace = "";
            if (context.type == Components.interfaces.koIFormatterContext.FORMAT_SELECTION) {
                // Remember the preceeding indentation.
                var match = context.text.match(/^(\s+)/);
                if (match) {
                    perLineWhitespace = match[1];
                }
            }
            // Remember the original settings, so we can compare to later.
            var pretext = context.text;
            var scimoz = view.scimoz;
            var original_currentPos = scimoz.currentPos;
            var original_anchor = scimoz.anchor;
            var original_line = scimoz.lineFromPosition(original_currentPos);
            var original_lineCount = scimoz.lineCount;

            try {
                formatter.format(context);
            } catch (ex) {
                ko.statusBar.AddMessage(ex, "formatting", 5000, true);
                return;
            }
            var text = context.text;
            // Formatting worked, update the text.
            if (context.type == Components.interfaces.koIFormatterContext.FORMAT_SELECTION) {
                var selPos = Math.min(original_currentPos, original_anchor);
                var startLinePos = scimoz.positionFromLine(scimoz.lineFromPosition(selPos));
                if (startLinePos < selPos) {
                    // Selection does not begin at the start of the line.
                    var preceedingText = scimoz.getTextRange(startLinePos, selPos);
                    var match = preceedingText.match(/^(\s+)/);
                    if (match && (match[1] == preceedingText)) {
                        // The preceeding text is just whitespace, need to
                        // maintain the indentation here as well.
                        preceedingSelectionWhitespace = preceedingText;
                        firstLineWhitespace = perLineWhitespace;
                        perLineWhitespace = preceedingText + perLineWhitespace;
                    }
                }
                if (perLineWhitespace) {
                    // Update all lines to include this preceeding whitespace.
                    var lines = text.split(/\r\n|\n|\r/g);
                    for (var i=0; i < lines.length; i++) {
                        /* Only indent lines that have some text. */
                        if (lines[i]) {
                            if (i == 0 && preceedingSelectionWhitespace) {
                                // The first line already has some preceeding
                                // whitespace from outside the user's selection.
                                // Only need to add in part of the whitespace.
                                lines[i] = firstLineWhitespace + lines[i];
                            } else {
                                lines[i] = perLineWhitespace + lines[i];
                            }
                        }
                    }

                    // Re-piece together the text.
                    text = lines.join("\n");

                    // Ensure the last line EOL state does not change.
                    var pretextEndsWithNewLine = pretext[pretext.length - 1].match(/\r|\n/);
                    var textEndsWithNewLine = text[text.length - 1].match(/\r|\n/);
                    if (!pretextEndsWithNewLine && textEndsWithNewLine) {
                        // The original text did not end with a newline, so
                        // the resulting text should not either.
                        text = text.substr(0, text.length - 1);
                    } else if (pretextEndsWithNewLine && !textEndsWithNewLine) {
                        // The original text ended with a newline, so the
                        // resulting text should as well.
                        text += "\n";
                    }
                }
            } else {
                // All the text is going to get replaced by insertText().
                scimoz.selectAll();
            }
            view.insertText(text);

            // Restore the line and possiblity the position as best as possible.
            var new_lineCount = scimoz.lineCount;
            var line_difference = new_lineCount - original_lineCount;
            if (context.type == Components.interfaces.koIFormatterContext.FORMAT_SELECTION) {
                // This will maintain the selection over the formatted area.
                if (original_currentPos <= original_anchor) {
                    scimoz.currentPos = original_currentPos;
                } else {
                    scimoz.anchor = original_anchor;
                }
            } else {
                // Just maintain the same line number, which may or may not be
                // the same line after the replacement.
                var new_line = original_line + line_difference;
                if (new_line < 0) {
                    new_line = 0;
                }
                var new_currentPos = scimoz.positionFromLine(new_line);
                scimoz.currentPos = new_currentPos;
                scimoz.anchor = new_currentPos;
            }

        } catch (ex) {
            log.exception(ex);
        }
    };

    /**
     * Create menuitems from the list of formatter configurations, which is a
     * list of koIPreferenceSet's. The menuitems are added to the provided
     * menupopup.
     */
    function _createFormatterMenuitemsFromFormatterConfigurations(menupopup, formatter_configurations) {
        var formatterSvc = Components.classes["@activestate.com/koFormatterService;1"].
                            getService(Components.interfaces.koIFormatterService);
        var pref;
        var name;
        var menuitem;
        var formatter;
        var formatter_name;
        for (i=0; i < formatter_configurations.length; i++) {
            pref = formatter_configurations[i];
            name = pref.getStringPref("name");
            formatter_name = pref.getStringPref("formatter_name");
            formatter = formatterSvc.getFormatterWithName(formatter_name);
            if (formatter) {
                menuitem = document.createElementNS(XUL_NS, 'menuitem');
                menuitem.setAttribute("id", "formatter_" + name);
                menuitem.setAttribute("dynamically_added_formatter", "true");
                menuitem.setAttribute('label', name);
                menuitem.setAttribute('tooltiptext', formatter.prettyName + " (" + pref.getStringPref("lang") + ")");
                menuitem.setAttribute('oncommand', 'ko.formatters.formatViewSelectionUsingConfigUuid("' + pref.getStringPref("uuid") + '");');
                menupopup.appendChild(menuitem);
            } else {
                this.log.warn("No formatter exists with the name: '" + formatter_name +
                              "', from preference '" + name + "'");
            }
        }
    };

    /**
     * Load the applicable formatter configurations into the menupopup, using
     * the currentView as the basis for determining the configuration language
     * and preferences.
     */
    function _createFormatterMenusForCurrentView(menupopup) {
        var view = ko.views.manager.currentView;
        if (!view) {
            return;
        }
        // Work out the applicable formatter configurations.
        var all_formatter_configurations = ko.formatters.getAllFormatterConfigurationsForPrefset(view.prefs);
        var lang_formatter_configurations = ko.formatters.getFormatterConfigurationsForView();
        var other_formatter_configurations = [];
        for (var i=0; i < all_formatter_configurations.length; i++) {
            if (lang_formatter_configurations.indexOf(all_formatter_configurations[i]) < 0) {
                other_formatter_configurations.push(all_formatter_configurations[i]);
            }
        }

        /* Remove any existing formatter menuitems. */
        var node = menupopup.firstChild;
        var nextSibling;
        while (node) {
            nextSibling = node.nextSibling;
            /* Only remove the menus that were previously added by this code. */
            /* Allows an overlay to add whatever they wish to this menu. */
            if (node.getAttribute("dynamically_added_formatter") == "true") {
                menupopup.removeChild(node);
            }
            node = nextSibling;
        }

        /* Add in formatter menus/menuitems for the formatter configurations. */
        _createFormatterMenuitemsFromFormatterConfigurations(menupopup, lang_formatter_configurations);
        /* Add in other formatter menus/menuitems. */
        if (other_formatter_configurations.length > 0) {
            if (lang_formatter_configurations.length > 0) {
                var menusep = document.createElementNS(XUL_NS, 'menuseparator');
                menusep.setAttribute("id", "editor_contextmenu_formatter_other_sep");
                menusep.setAttribute("dynamically_added_formatter", "true");
                menupopup.appendChild(menusep);
            }
            var other_menupopup = document.createElementNS(XUL_NS, 'menupopup');
            var other_menu = document.createElementNS(XUL_NS, 'menu');
            var strbundle = document.getElementById("formatter_strings");
            other_menu.setAttribute("label", strbundle.getString("otherFormattersMenulistLabel"));
            other_menu.setAttribute("dynamically_added_formatter", "true");
            var other_menupopup = document.createElementNS(XUL_NS, 'menupopup');
            other_menu.appendChild(other_menupopup);
            _createFormatterMenuitemsFromFormatterConfigurations(other_menupopup, other_formatter_configurations);
            menupopup.appendChild(other_menu);
        }
    };


    /************************************************************
     *        Exposed ko.formatters.XYZ functions               *
     ************************************************************/

    /**
     * Format this view. If there is a selection in this view or the
     * formatSelection argument is true, then the formatter will work upon the
     * view's selection, else the formatter works on the complete document.
     * The targetted text will be replaced if the formatting is successful.
     * @param view {Components.interfaces.koIScintillaView}
     *        (Optional) The view to format. Defaults to the currentview.
     * @param formatSelection {boolean}
     *        (Optional) Force the formater to work on a selection of the view.
     */
    this.formatView = function formatView(view, formatSelection /* false */) {
        try {
            if (!view) {
                view = ko.views.manager.currentView;
            }
            if (!view || (view.getAttribute('type') != 'editor')) {
                return;
            }
            var formatterSvc = Components.classes["@activestate.com/koFormatterService;1"].
                                getService(Components.interfaces.koIFormatterService);
            var formatter_out = {};
            var context_out = {};
            formatterSvc.getFormatterAndContextFromView(view, formatter_out, context_out);
            var formatter = formatter_out.value;
            var context = context_out.value;
            try {
                if (formatSelection && view.scimoz.selText) {
                    // Ensure the context is the selection.
                    context.type = Components.interfaces.koIFormatterContext.FORMAT_SELECTION;
                    context.text = view.scimoz.selText;
                }
                if (formatter) {
                    _formatViewWithFormatterAndContext(view, formatter, context);
                } else {
                    ko.statusBar.AddMessage("No formatter configured for language: " +
                                            view.document.language,
                                            "formatting", 5000, true);
                }
            } finally {
                // Ensure the context is cleaned-up.
                context.text = "";
                context.scimoz = null;
                context.prefset = null;
                context.formatter_prefset = null;
            }
        } catch (ex) {
            log.exception(ex);
        }
    }

    /**
     * Return a list of all formatter configuration preferences.
     * @returns {array}  List of koIPreferenceSet xpcom objects.
     */
    this.getAllFormatterConfigurationsForPrefset = function getAllFormatterConfigurationsForPrefset(prefset) {
        var configuration_prefsets = [];
        if (prefset.hasPref('configuredFormatters')) {
            var prefs = prefset.getPref('configuredFormatters');
            for (var i=0; i < prefs.length; i++) {
                configuration_prefsets.push(prefs.getPref(i));
            }
        }
        return configuration_prefsets;
    }

    /**
     * Return a list of formatter configuration preferences that are setup for
     * the given language.
     * @returns {array}  List of koIPreferenceSet xpcom objects.
     */
    this.getFormatterConfigurationsForLangAndPrefset = function getFormatterConfigurationsForLangAndPrefset(lang, prefset) {
        var all_configuration_prefsets = this.getAllFormatterConfigurationsForPrefset(prefset);
        var lang_configuration_prefsets = [];
        for (var i=0; i < all_configuration_prefsets.length; i++) {
            if (all_configuration_prefsets[i].getStringPref("lang") == lang) {
                lang_configuration_prefsets.push(all_configuration_prefsets[i]);
            }
        }
        return lang_configuration_prefsets;
    }

    /**
     * Return a list of formatter configuration preferences that are setup for
     * the current view's language and prefset.
     * @returns {array}  List of koIPreferenceSet xpcom objects.
     */
    this.getFormatterConfigurationsForView = function getFormatterConfigurationsForView(view) {
        if (!view) {
            view = ko.views.manager.currentView;
            if (!view) {
                return null;
            }
        }
        return this.getFormatterConfigurationsForLangAndPrefset(view.document.language, view.prefs);
    }

    /**
     * Populate the provided menupopup element with the available formatters
     * for the current view.
     */
    this.loadFormatterMenusForCurrentView = function loadFormatterMenusForCurrentView(menupopup, event) {
        try {
            if (event.target && event.target.getAttribute("id") != menupopup.getAttribute("id")) {
                // Avoid loading sub-menu popups, as they are already loaded.
                return;
            }
            _createFormatterMenusForCurrentView(menupopup);
        } catch (ex) {
            log.exception(ex);
        }
    }

    /**
     * Format the given view (defaults to the current view when not supplied)
     * with the formatter configuration provided. If there is a selection in
     * this view then the formatter will work upon the view's selection, else
     * the formatter works on the complete document. The targetted text will be
     * replaced if the formatting is successful.
     * 
     * @param view {Components.interfaces.koIScintillaView}
     *        The view to format. Defaults to the current view if null.
     * @param config {Components.interfaces.koIPreferenceSet}
     *        The formatter configuration prefset to use for formatting.
     * @param formatSelection {boolean}
     *        (Optional) Force the formater to work on a selection of the view.
     */
    this.formatViewUsingConfig = function formatViewUsingConfig(view, config, formatSelection) {
        log.debug("formatting view using: " + config.getStringPref("name"));
        var formatterSvc = Components.classes["@activestate.com/koFormatterService;1"].
                            getService(Components.interfaces.koIFormatterService);
        var formatter = formatterSvc.getFormatterWithName(config.getStringPref("formatter_name"));
        if (formatter) {
            var context = formatterSvc.createFormatterContextFromView(view);
            try {
                if (formatSelection &&
                    (context.type != Components.interfaces.koIFormatterContext.FORMAT_SELECTION)) {
                    var selText = view.scimoz.selText;
                    if (selText) {
                        context.type = Components.interfaces.koIFormatterContext.FORMAT_SELECTION;
                        context.text = selText;
                    }
                }
                context.formatter_prefset = config;
                _formatViewWithFormatterAndContext(view, formatter, context);
            } finally {
                // Ensure the context instance is cleaned-up.
                context.text = "";
                context.prefset = null;
                context.formatter_prefset = null;
                context.scimoz = null;
            }
        } else {
            ko.dialogs.alert("No formatter exists with the name: '" +
                             formatter_name + "', from preference '" +
                             name + "'",
                             null,
                             "Invalid Formatter Configuration");
        }
    }

    /**
     * Format the given view (defaults to the currentView when not supplied)
     * with the formatter configuration that has the supplied config name.
     * If there is a selection in this view then the formatter will work upon
     * the view's selection, else the formatter works on the complete document.
     * The targetted text will be replaced if the formatting is successful.
     
     * @param view {Components.interfaces.koIScintillaView}
     *        The view to format. Defaults to the current view if null.
     * @param configName {string}
     *        The name of the formatter configuration to use for formatting.
     */
    this.formatViewUsingConfigName = function formatViewUsingConfigName(view, configName) {
        try {
            if (!view) {
                view = ko.views.manager.currentView;
                if (!view) {
                    return;
                }
            }
            // Work out the applicable formatter configurations.
            var all_formatter_configurations = this.getAllFormatterConfigurationsForPrefset(view.prefs);
            for (var i=0; i < all_formatter_configurations.length; i++) {
                if (all_formatter_configurations[i].getStringPref("name") == configName) {
                    this.formatViewUsingConfig(view, all_formatter_configurations[i]);
                    break;
                }
            }
        } catch (ex) {
            log.exception(ex);
        }
    },

    /**
     * Format the current view with the formatter configuration that has the
     * provided UUID. This is required for the formatter menu(item)'s.
     * 
     * @param uuid {string}  The unique identifier of the configuration prefset.
     */
    this.formatViewSelectionUsingConfigUuid = function formatViewSelectionUsingConfigUuid(uuid) {
        try {
            var view = ko.views.manager.currentView;
            if (!view) {
                return;
            }
            // Work out the applicable formatter configurations.
            var all_formatter_configurations = this.getAllFormatterConfigurationsForPrefset(view.prefs);
            for (var i=0; i < all_formatter_configurations.length; i++) {
                if (all_formatter_configurations[i].getStringPref("uuid") == uuid) {
                    this.formatViewUsingConfig(view, all_formatter_configurations[i]);
                    break;
                }
            }
        } catch (ex) {
            log.exception(ex);
        }
    }

    // Controller instance on the current window to handle formatter commands.
    window.controllers.appendController(new FormattersController());

}).apply(ko.formatters);
